/*
Capture patch <http://quakecapture.sourceforge.net/>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <string.h>
#include <stdlib.h>

#include "patch.h"

// API to access implementations of capture
#include "bridge.h"

// Declarations of all necessary Quake engine glue
#include "helpers.h"


// My private variables for buffering audio
short capture_audio_samples[44100]; // big enough buffer for 1fps at 44100KHz
int captured_audio_samples;


/*  My private helpers
 */
int Cap_AudioSamplesPerVideoFrame(void)
{
    return (int)(0.5 + (float)CaptureHelper_AudioHz() / CaptureHelper_VideoFps());
}

void Cap_Start(char* filename)
{
    int success = 0;

    COM_StripExtension(filename, filename);

    success = Capture_Open(
        Cvar_VariableString("capture_mode"),
        filename,
        Cvar_VariableString("capture_codec"),
        CaptureHelper_VideoFps(),
        CaptureHelper_AudioHz()
    );

    if (success && Capture_IsRecording()) {
        CaptureHelper_Printf_s("Capturing to %s\n", filename);
    } else {
        CaptureHelper_Print(Capture_Error());
        CaptureHelper_Print("Failed to open capture.\n");
    }
}

void Cap_CaptureFps_f(void)
{
    if (Cmd_Argc() == 1) {
        CaptureHelper_Printf_f("\"capture_fps\" is \"%.3f\"\n", CaptureHelper_VideoFps());
    } else {
        CaptureHelper_SetVideoFps((float)atof(Cmd_Argv(1)));
    }
}

void Cap_Start_f(void)
{
    if (Cmd_Argc() != 2) {
		CaptureHelper_Print("capture_start <filename>: Start capturing to named file.\n");
		return;
	}

    if (!CaptureHelper_VideoFps()) {
        CaptureHelper_Print("Set capture_fps before starting capture.\n");
        return;
    }

    Cap_Start(Cmd_Argv(1));
}

void Cap_Stop_f(void)
{
    Capture_Close();
    CaptureHelper_Print("Stopped capturing.\n");
}

void Cap_CaptureDemo_f(void)
{
    if (Cmd_Argc() != 2) {
		CaptureHelper_Printf_s(
            "capture_demo <demoname> : capture <demoname>%s then exit\n",
            CaptureHelper_DefaultDemoExtension()
        );
		return;
	}

	CaptureHelper_Printf_s("Starting capture of %s\n", Cmd_Argv(1));

    if (!CaptureHelper_VideoFps()) {
        CaptureHelper_SetVideoFps(15.0);
    }

    if (CaptureHelper_PlayDemo_f()) {
        Cap_Start(COM_SkipPath(Cmd_Argv(1)));
        CaptureHelper_SetBatchCapture(Capture_IsRecording());
    }
}


/*  Implementation of the patch API
 */
void CapturePatch_Init(void)
{
    captured_audio_samples = 0;

    // Interactively start capture of current world
    Cmd_AddCommand("capture_start", Cap_Start_f);

    // Interactively finish the current capture
    Cmd_AddCommand("capture_stop", Cap_Stop_f);

    // Capture all of an existing demo and exit on completion.
    // (Designed for batch processing.)
    Cmd_AddCommand("capture_demo", Cap_CaptureDemo_f);

    // A fake variable that controls the capture framerate
    Cmd_AddCommand("capture_fps", Cap_CaptureFps_f);

    // Mode of capture
    CaptureHelper_InitVariable("capture_mode", "avi");

    // Codec to use for compressing captures. "0" indicates no codec.
    CaptureHelper_InitVariable("capture_codec", "0");
}

void CapturePatch_OnStopPlayback(void)
{
    if (!CaptureHelper_BatchCapture()) return;

	CaptureHelper_SetBatchCapture(0);
    Cap_Stop_f();

    CaptureHelper_Shutdown();
}

void CapturePatch_OnUpdateScreen(void)
{
    unsigned char* buffer;
    int width = CaptureHelper_VideoWidth();
    int height = CaptureHelper_VideoHeight();
    int size = width * height * 3;
    int success = 0;

    if (!CapturePatch_IsActive()) return;

    buffer = malloc(size);

    if (CaptureHelper_CopyVideo(buffer)) {
        success = Capture_WriteVideo(width, height, buffer);
    }

    if (!success) {
        CaptureHelper_Print("Problem capturing video frame!\n");
        CaptureHelper_SetBatchCapture(0); // don't exit when a capture_demo ends
    }
    free(buffer);
}

void CapturePatch_OnTransferStereo16(void)
{
    if (!CapturePatch_IsActive()) return;

    // Copy last audio chunk written into our temporary buffer
    memcpy(
        capture_audio_samples + 2 * captured_audio_samples,
        snd_out,
        snd_linear_count * CaptureHelper_SoundChannels()
    );
    captured_audio_samples += snd_linear_count / 2;

    if (captured_audio_samples >= Cap_AudioSamplesPerVideoFrame()) {

        if (!Capture_WriteAudio(captured_audio_samples, (unsigned char*)capture_audio_samples)) {
            CaptureHelper_Print("Problem capturing audio!\n");
            CaptureHelper_SetBatchCapture(0); // don't exit when a capture_demo ends  /** **/
        }
        captured_audio_samples = 0;

    }
}

int CapturePatch_GetSoundtime(void)
{
    if (CapturePatch_IsActive()) {
        soundtime += Cap_AudioSamplesPerVideoFrame();
        return 1;
    } else {
        return 0;
    }
}

int CapturePatch_IsActive(void)
{
    if (CaptureHelper_ShouldSuppressCapture()) return 0;

    // otherwise output if a file is open to write to
    return Capture_IsRecording();
}